@@ -26,7 +26,6 @@ from .const import (
     Commands,
     DoorStatus,
     LockInfo,
-    LockState,
     LockStatus,
 )
 from .secure_session import SecureSession
@@ -172,8 +171,14 @@ class Lock:
         # Order matters here, we must start notify for the secure session before
         # the non-secure session
         await self.secure_session.start_notify()
+        try:
+            await self._setup_session()
+        except BleakError as err:
+            error_desc = str(err).lower()
+            if "invalid handle" in error_desc:
+                await self._handle_missing_characteristic(error_desc)
+            raise err
         await self.session.start_notify()
-        await self._setup_session()
 
     async def _handle_missing_characteristic(self, char_uuid: str) -> None:
         """Handle missing characteristic."""
@@ -209,7 +214,7 @@ class Lock:
         elif state[0] == 0xAA:
             if state[1] == Commands.UNLOCK.value:
                 self._state_callback([LockStatus.UNLOCKED])
-            if state[1] == Commands.LOCK.value:
+            elif state[1] == Commands.LOCK.value:
                 self._state_callback([LockStatus.LOCKED])
             else:
                 _LOGGER.debug("%s: Unknown state: %s", self.name, state.hex())
@@ -226,7 +231,9 @@ class Lock:
         # Send SEC_LOCK_TO_MOBILE_KEY_EXCHANGE
         cmd = self.secure_session.build_command(0x01)
         util._copy(cmd, handshake_keys[0x00:0x08], destLocation=0x04)
-        response = await self.secure_session.execute(cmd)
+        response = await self.secure_session.execute(
+            cmd, "SEC_LOCK_TO_MOBILE_KEY_EXCHANGE"
+        )
         if response[0x00] != 0x02:
             raise AuthError(
                 "Authentication error: key or slot (key index) is incorrect: unexpected response to SEC_LOCK_TO_MOBILE_KEY_EXCHANGE: "
@@ -243,13 +250,15 @@ class Lock:
         # Send SEC_INITIALIZATION_COMMAND
         cmd = self.secure_session.build_command(0x03)
         util._copy(cmd, handshake_keys[0x08:0x10], destLocation=0x04)
-        response = await self.secure_session.execute(cmd)
+        response = await self.secure_session.execute(cmd, "SEC_INITIALIZATION_COMMAND")
         if response[0] != 0x04:
             raise AuthError(
                 "Authentication error: key or slot (key index) is incorrect: unexpected response to SEC_INITIALIZATION_COMMAND: "
                 + response.hex()
             )
         self.session.set_key(session_key)
+        self.secure_session.enable_cooldown()
+        self.session.enable_cooldown()
 
     @raise_if_not_connected
     async def lock_info(self) -> LockInfo:
@@ -277,7 +286,9 @@ class Lock:
         """Force the lock to lock."""
         _LOGGER.debug("%s: Locking", self.name)
         assert self.session is not None  # nosec
-        await self.session.execute(self.session.build_command(Commands.LOCK.value))
+        await self.session.execute(
+            self.session.build_command(Commands.LOCK.value), "force_lock"
+        )
         _LOGGER.debug("%s: Finished locking", self.name)
 
     @raise_if_not_connected
@@ -285,52 +296,70 @@ class Lock:
         """Force the lock to unlock."""
         _LOGGER.debug("%s: Unlocking", self.name)
         assert self.session is not None  # nosec
-        await self.session.execute(self.session.build_command(Commands.UNLOCK.value))
+        await self.session.execute(
+            self.session.build_command(Commands.UNLOCK.value), "force_unlock"
+        )
         _LOGGER.debug("%s: Finished unlocking", self.name)
 
     async def lock(self) -> None:
-        if (await self.status()).lock == LockStatus.UNLOCKED:
+        if (await self.lock_status()) == LockStatus.UNLOCKED:
             await self.force_lock()
 
     async def unlock(self) -> None:
-        if (await self.status()).lock == LockStatus.LOCKED:
+        if (await self.lock_status()) == LockStatus.LOCKED:
             await self.force_unlock()
 
-    async def _execute_command(self, cmd_byte: int) -> bytes:
+    async def _execute_command(self, cmd_byte: int, command_name: str) -> bytes:
         assert self.session is not None  # nosec
-        response = await self.session.execute(
-            self.session.build_operation_command(cmd_byte)
+        command = self.session.build_operation_command(cmd_byte)
+        _LOGGER.debug("%s: send: [%s] [%s]", self.name, command.hex(), hex(cmd_byte))
+        response = await self.session.execute(command, command_name)
+        _LOGGER.debug(
+            "%s: response: [%s] [%s]", self.name, response.hex(), hex(cmd_byte)
         )
-        _LOGGER.debug("%s: response: [%s]", self.name, response.hex())
         return response
 
     def _parse_lock_and_door_state(
         self, response: bytes
     ) -> tuple[LockStatus, DoorStatus]:
         """Parse the lock and door state from the response."""
-        lock_status = response[0x08]
-        door_status = response[0x09]
-        lock_status_enum = VALUE_TO_LOCK_STATUS.get(lock_status, LockStatus.UNKNOWN)
-        door_status_enum = VALUE_TO_DOOR_STATUS.get(door_status, DoorStatus.UNKNOWN)
+        return self._parse_lock_status(response[0x08]), self._parse_door_status(
+            response[0x09]
+        )
 
+    def _parse_lock_status(self, lock_status: int) -> LockStatus:
+        """Parse the lock state from the response."""
+        lock_status_enum = VALUE_TO_LOCK_STATUS.get(lock_status, LockStatus.UNKNOWN)
         if lock_status_enum == LockStatus.UNKNOWN:
             _LOGGER.debug(
                 "%s: Unrecognized lock_status_str code: %s", self.name, hex(lock_status)
             )
+        return lock_status_enum
+
+    def _parse_door_status(self, door_status: int) -> DoorStatus:
+        """Parse the door state from the response."""
+        door_status_enum = VALUE_TO_DOOR_STATUS.get(door_status, DoorStatus.UNKNOWN)
         if door_status_enum == DoorStatus.UNKNOWN:
             _LOGGER.debug(
                 "%s: Unrecognized door_status_str code: %s", self.name, hex(door_status)
             )
-        return lock_status_enum, door_status_enum
+        return door_status_enum
 
     @raise_if_not_connected
-    async def status(self) -> LockState:
-        _LOGGER.debug("%s: Executing status", self.name)
-        response = await self._execute_command(
-            0x2F if self._lock_info and self._lock_info.door_sense else 0x02
-        )
-        _LOGGER.debug("%s: Finished executing status", self.name)
-        return LockState(*self._parse_lock_and_door_state(response), None, None)
+    async def lock_status(self) -> LockStatus:
+        _LOGGER.debug("%s: Executing lock_status", self.name)
+        # We used to use 0x2F here but it seems to be broken on some locks
+        response = await self._execute_command(0x02, "lock_status")
+        _LOGGER.debug("%s: Finished executing lock_status", self.name)
+        return self._parse_lock_status(response[0x08])
+
+    @raise_if_not_connected
+    async def door_status(self) -> DoorStatus:
+        _LOGGER.debug("%s: Executing door_status", self.name)
+        # We used to use 0x2F here but it seems to be broken on some locks
+        response = await self._execute_command(0x2E, "door_status")
+        _LOGGER.debug("%s: Finished executing door_status", self.name)
+        return self._parse_door_status(response[0x08])
 
     def _parse_battery_state(self, response: bytes) -> BatteryState:
         """Parse the battery state from the response."""
@@ -346,7 +375,7 @@ class Lock:
     @raise_if_not_connected
     async def battery(self) -> BatteryState:
         _LOGGER.debug("%s: Executing battery", self.name)
-        response = await self._execute_command(0x0F)
+        response = await self._execute_command(0x0F, "battery")
         _LOGGER.debug("%s: Finished executing battery", self.name)
         return self._parse_battery_state(response)
 
@@ -364,6 +393,8 @@ class Lock:
     async def _shutdown_connection(self) -> None:
         """Shutdown the connection."""
         _LOGGER.debug("%s: Shutting down the connection", self.name)
+        if self.session:
+            await self.session.stop_notify()
         if (
             not self.is_secure
             or not self.secure_session
@@ -374,7 +405,8 @@ class Lock:
         cmd[0x11] = 0x00
         response = None
         try:
-            response = await self.secure_session.execute(cmd)
+            response = await self.secure_session.execute(cmd, "shutdown")
+            await self.secure_session.stop_notify()
         except (AuthError, DisconnectedError):
             # Lock already disconnected us
             return
