@@ -54,9 +54,13 @@ DISCONNECT_DELAY = 5.1
 
 RESYNC_DELAY = 0.01
 
+# Number of seconds to wait after the first connection
+# to disconnect to free up the bluetooth adapter.
+FIRST_CONNECTION_DISCONNECT_TIME = 2.1
+
 # After a lock operation we need to wait for the lock to
 # update its state or it will return a stale state.
-LOCK_STALE_STATE_DEBOUNCE_DELAY = 4.0
+LOCK_STALE_STATE_DEBOUNCE_DELAY = 6.1
 
 # How long to wait before processing an advertisement change
 ADV_UPDATE_COALESCE_SECONDS = 0.05
@@ -131,6 +135,7 @@ def retry_bluetooth_connection_error(func: WrapFuncType) -> WrapFuncType:
                 # point in retrying.
                 raise
             except RETRY_BACKOFF_EXCEPTIONS as err:
+                await self._async_handle_disconnected(err)
                 if attempt >= max_attempts:
                     _LOGGER.debug(
                         "%s: %s error calling %s, reach max attempts (%s/%s)",
@@ -156,6 +161,7 @@ def retry_bluetooth_connection_error(func: WrapFuncType) -> WrapFuncType:
                 )
                 await asyncio.sleep(0.25)
             except RETRY_EXCEPTIONS as err:
+                await self._async_handle_disconnected(err)
                 if attempt >= max_attempts:
                     _LOGGER.debug(
                         "%s: %s error calling %s, reach max attempts (%s/%s)",
@@ -223,8 +229,12 @@ class PushLock:
         self._cancel_deferred_update: asyncio.TimerHandle | None = None
         self._client: Lock | None = None
         self._connect_lock = asyncio.Lock()
+        self._seen_this_session: set[
+            type[LockStatus] | type[DoorStatus] | type[BatteryState] | type[AuthState]
+        ] = set()
         self._disconnect_timer: asyncio.TimerHandle | None = None
         self._idle_disconnect_delay = idle_disconnect_delay
+        self._next_disconnect_delay = idle_disconnect_delay
         self._first_update_future: asyncio.Future[None] | None = None
         self._background_tasks: set[asyncio.Task[None]] = set()
         self._auth_failures: int = 0
@@ -343,17 +353,18 @@ class PushLock:
         """Reset disconnect timer."""
         self._cancel_disconnect_timer()
         self._expected_disconnect = False
+        timeout = self._next_disconnect_delay
+        _LOGGER.debug(
+            "%s: Resetting disconnect timer to %s seconds", self.name, timeout
+        )
         self._disconnect_timer = self.loop.call_later(
-            self._idle_disconnect_delay, self._disconnect_with_timer
+            timeout, self._disconnect_with_timer, timeout
         )
 
-    async def _execute_forced_disconnect(self) -> None:
+    async def _execute_forced_disconnect(self, reason: str) -> None:
         """Execute forced disconnection."""
         self._cancel_disconnect_timer()
-        _LOGGER.debug(
-            "%s: Executing forced disconnect",
-            self.name,
-        )
+        _LOGGER.debug("%s: Executing forced disconnect: %s", self.name, reason)
         if (update_task := self._update_task) and not update_task.done():
             self._update_task = None
             update_task.cancel()
@@ -361,7 +372,7 @@ class PushLock:
                 await update_task
         await self._execute_disconnect()
 
-    def _disconnect_with_timer(self) -> None:
+    def _disconnect_with_timer(self, timeout: float) -> None:
         """Disconnect from device."""
         if self._operation_lock.locked():
             _LOGGER.debug("%s: Disconnect timer reset due to operation lock", self.name)
@@ -376,7 +387,7 @@ class PushLock:
             self._deferred_update()
             return
         self._cancel_disconnect_timer()
-        self.background_task(self._execute_timed_disconnect())
+        self.background_task(self._execute_timed_disconnect(timeout))
 
     def _cancel_disconnect_timer(self) -> None:
         """Cancel disconnect timer."""
@@ -384,15 +395,27 @@ class PushLock:
             self._disconnect_timer.cancel()
             self._disconnect_timer = None
 
-    async def _execute_timed_disconnect(self) -> None:
+    async def _execute_timed_disconnect(self, timeout: float) -> None:
         """Execute timed disconnection."""
         _LOGGER.debug(
             "%s: Executing timed disconnect after timeout of %s",
             self.name,
-            DISCONNECT_DELAY,
+            timeout,
         )
         await self._execute_disconnect()
 
+    async def _async_handle_disconnected(self, exc: Exception) -> None:
+        """Clean up after a disconnect."""
+        _LOGGER.debug("%s: Disconnected due to %s, cleaning up", self.name, exc)
+        if self._connect_lock.locked():
+            _LOGGER.error(
+                "%s: Disconnected while connection was in progress, ignoring",
+                self.name,
+            )
+            return
+        self._cancel_disconnect_timer()
+        await self._execute_disconnect()
+
     async def _execute_disconnect(self) -> None:
         """Execute disconnection."""
         async with self._connect_lock:
@@ -430,7 +453,9 @@ class PushLock:
                 )
                 await self._client.disconnect()
                 raise
+            self._next_disconnect_delay = self._idle_disconnect_delay
             self._reset_disconnect_timer()
+            self._seen_this_session.clear()
             return self._client
 
     async def lock(self) -> None:
@@ -477,15 +502,21 @@ class PushLock:
         self._reset_disconnect_timer()
         self._update_any_state(states)
 
+    def _get_current_state(self) -> LockState:
+        """Get the current state of the lock."""
+        return self._lock_state or LockState(
+            self.lock_status, self.door_status, self.battery, self.auth
+        )
+
     def _update_any_state(
         self, states: Iterable[LockStatus | DoorStatus | BatteryState | AuthState]
     ) -> None:
         _LOGGER.debug("%s: State changed: %s", self.name, states)
-        lock_state = self._lock_state or LockState(
-            self.lock_status, self.door_status, self.battery, self.auth
-        )
+        lock_state = self._get_current_state()
         original_lock_status = lock_state.lock
         for state in states:
+            state_type = type(state)
+            self._seen_this_session.add(state_type)
             if isinstance(state, AuthState):
                 lock_state = replace(lock_state, auth=state)
             elif isinstance(state, LockStatus):
@@ -536,33 +567,62 @@ class PushLock:
         _LOGGER.debug(
             "%s: Starting update (has_lock_info: %s)", self.name, has_lock_info
         )
-        try:
-            lock = await self._ensure_connected()
-            if not self._lock_info:
-                self._lock_info = await lock.lock_info()
-            state = await lock.status()
+        lock = await self._ensure_connected()
+        if not self._lock_info:
+            self._lock_info = await lock.lock_info()
+        # Asking for battery first seems to be reduce the chance of the lock
+        # getting into a bad state.
+        state = self._get_current_state()
+        if type(BatteryState) not in self._seen_this_session:
             battery_state = await lock.battery()
-        except asyncio.CancelledError:
-            _LOGGER.debug(
-                "%s: In-progress update canceled due "
-                "to lock operation or setup timeout",
-                self.name,
-            )
-            raise
-        else:
             self._auth_failures = 0
             state = replace(
                 state, battery=battery_state, auth=AuthState(successful=True)
             )
+        # Only ask for the lock status if we haven't seen
+        # it this session since notify callbacks will happen
+        # if it changes and the extra polling can cause the lock
+        # to get into a bad state.
+        if type(LockStatus) not in self._seen_this_session:
+            lock_status = await lock.lock_status()
+            self._auth_failures = 0
+            state = replace(state, lock=lock_status, auth=AuthState(successful=True))
+        if (
+            type(DoorStatus) not in self._seen_this_session
+            and self._lock_info
+            and self._lock_info.door_sense
+        ):
+            door_status = await lock.door_status()
+            self._auth_failures = 0
+            state = replace(state, door=door_status, auth=AuthState(successful=True))
         _LOGGER.debug("%s: Finished update", self.name)
         self._callback_state(state)
 
+        if state.battery and state.battery.voltage <= 3.0:
+            _LOGGER.debug(
+                "%s: Battery voltage is impossible: %s",
+                self.name,
+                state.battery.voltage,
+            )
+            # If the battery voltage is impossible, reconnect.
+            await self._execute_forced_disconnect("impossible battery voltage")
+
+        if state.lock in (LockStatus.UNKNOWN_01, LockStatus.UNKNOWN_06):
+            _LOGGER.debug("%s: Lock is in an unknown state: %s", self.name, state.lock)
+            # If the lock is in a bad state, reconnect.
+            await self._execute_forced_disconnect(
+                f"lock is in unknown state: {state.lock}"
+            )
+
         if not has_lock_info:
             # On first update free up the connection
             # so we can bring other locks online if
             # the bluetooth adapter is out of connections
-            # slots
-            await self._execute_forced_disconnect()
+            # slots. We reset the timer to a low number
+            # so that if another update request is pending
+            # we do not disconnect until it completes.
+            self._next_disconnect_delay = FIRST_CONNECTION_DISCONNECT_TIME
+            self._reset_disconnect_timer()
 
         return state
 
@@ -678,7 +738,7 @@ class PushLock:
         def _cancel() -> None:
             self._running = False
             self._cancel_future_update()
-            self.background_task(self._execute_forced_disconnect())
+            self.background_task(self._execute_forced_disconnect("stopping"))
 
         return _cancel
 
@@ -796,6 +856,7 @@ class PushLock:
         except asyncio.CancelledError:
             self._set_update_state(RuntimeError("Update was canceled"))
             _LOGGER.debug("%s: In-progress update canceled", self.name)
+            raise
         except asyncio.TimeoutError as ex:
             self._set_update_state(ex)
             _LOGGER.exception("%s: Timed out updating", self.name)
@@ -804,6 +865,11 @@ class PushLock:
             wrapped_bleak_exc.__cause__ = ex
             self._set_update_state(wrapped_bleak_exc)
             _LOGGER.exception("%s: Bluetooth error updating", self.name)
+        except DisconnectedError as ex:
+            wrapped_bleak_exc = BluetoothError(str(ex))
+            wrapped_bleak_exc.__cause__ = ex
+            self._set_update_state(wrapped_bleak_exc)
+            _LOGGER.exception("%s: Disconnected while updating", self.name)
         except Exception as ex:  # pylint: disable=broad-except
             wrapped_exc = YaleXSBLEError(str(ex))
             wrapped_exc.__cause__ = ex
